; display.inc - Display routines
;
; Copyright (C) 1999 Thomas Novelli
;	Retro is distributed under the terms of the GNU LGPL, version 2 or
;	greater. (For the text of the license, see the file COPYING.LIB)
;
; Wish list:
;	Scrollback: in 'scroll', copy top line to a buffer.
;


	struc BITMAP
.w	resd	1	;size
.h	resd	1
.line	;; array of [.h] pointers to the beginning of each line
	endstruc


init_screen:
;
; Get cursor position from BIOS data area
;
	movzx	ax, byte [0x451]	;row
	mov	dx, 160
	mul	dx
	movzx	dx, byte [0x450]	;column
	shl	dx, 1
	add	ax, dx
	mov	[cursor], ax

	mov	al, 0x30
	and	al, [0x410]
	cmp	al, 0x30
	je	.mono
	;; setup for color
	jmp short .1
.mono	;; setup for mono
	mov dword [vidbuf], 0xB0000
	;different i/o port..
.1
	ret


putc:
EMIT:		;( char -- , display a character )
;
; Print a character & advance cursor
;
; To do:
;	More terminal handling stuff
;
	mpush	eax, ecx, edi
	DPOP	eax
	mov	ecx, .call-.chars
	mov	edi, .chars
	cld
	repne	scasb
	jne	.putc	;No match?
	neg	ecx
	add	ecx, .call-.chars-1
	call	[.call+ecx*4]
	jmp short .end
.putc	mov	ah, [attr]
	movzx	edi, word [cursor]
	add	edi, [vidbuf]
	stosw
	add word [cursor], 2

	call	scroll
	call	update_cursor
.end	mpop	eax, ecx, edi
	ret

.chars	db	0,	8,	9,	10,	127
.call	dd	.null,	.bs,	.tab,	_cr,	.del

.null	ret

.bs	sub word [cursor], 2
	movzx	edi, word [cursor]
	add	edi, [vidbuf]
	xor	al, al
	stosb
	call	update_cursor
	ret

.tab	;; round [cursor] up to the nearest 16
	and dword [cursor], ~0x0F
	add dword [cursor], 0x10
	ret

.del	ret
eEMIT



_cr:
;
; Affects AX, CX, DX
;
	mpush	eax, ecx, edx
	mov	ax, [cursor]
	xor	dx, dx
	mov	cx, 160
	div	cx		;dx = remainder
	mov	ax, 160
	sub	ax, dx
	add	[cursor], ax
	call	scroll
	call	update_cursor
	mpop	eax, ecx, edx
	ret



scroll:
;
; If the cursor is past the bottom of the screen, scroll down 1 line.
;
	cmp word [cursor], 80*25*2	;Need to scroll?
	jb	.end
	mpush	eax, ecx, esi, edi
	mov	edi, [vidbuf]		;Move top 24 rows up one
	mov	esi, edi
	add	esi, [cols]
	mov	ecx, 80*24*2
	cld
	rep	movsb

	mov	ah, [attr]		;Clear bottom row
	xor	al, al
	mov	ecx, 80
	rep	stosw
	mpop	eax, ecx, esi, edi
	sub word [cursor], 80*2		;Adjust cursor position
.end
	ret



print:	;; ( string -- )
;
; Print a null-terminated string
;
	mpush	eax, esi

	DPOP	esi
;	mov	esi, [esp+3*4]
	cld
.loop
	lodsb
	or	al, al			;0 = end
	jz	.end
	DPUSH	eax
	call	putc
	jmp short .loop

.end
	mpop	eax, esi
	ret



cprint:
;
; Print a counted string (1st byte is count)
;
	mpush	eax, ecx, esi
;	mov	esi, [esp+4*4]
	DPOP	esi
	cld
	lodsb
	movzx	ecx, al
.loop
	lodsb
	DPUSH	eax
	call	putc
	loop	.loop

	mpop	eax, ecx, esi
	ret



print_hex:
;
; Print an 8-digit hex number
;
	mpush	eax, ecx, edx
;	mov	edx, [esp+4*4]
	DPOP	edx
	mov	ecx, 8
.pdigit
	rol	edx, 4
	mov	ax, 0x0E0F
	and	al, dl
	add	al, 0x90
	daa
	adc	al, 0x40
	daa
	DPUSH	eax
	call	putc
	loop	.pdigit

	DPUSH	32
	call	putc
	mpop	eax, ecx, edx
	ret



%if 0
show_block:
;
; Quickie debugging routine. Good for FORTH, though...
;
;input: ESI
;
	mov	edi, [vidbuf]
	mov	ah, [attr]
	mov	ecx, 16		;lines
.l1
	push	ecx
	mov	ecx, 64
.l2
	lodsb
	stosw
	loop	.l2
	add	edi, (80-64)*2
	pop	ecx
	loop	.l1
	ret
%endif



%if 0
getxy:
	mov	ax, [cursor]
	xor	dx, dx
	div	dword [cols]
	shr	dx, 1
; ax=Y, dx=X
	ret
%endif



update_cursor:
;
; Physically move the cursor to [x,y], using port I/O
;
; Index port: 0x3d4
; Data port:  0x3d5
;
	mpush	eax, edx
	mov	al, 0x0f	;Index reg.: cursor pos (low byte)
	mov	dx, 0x3d4
	out	dx, al
	mov	ax, [cursor]	;Write low byte
	shr	ax, 1
	inc	edx		;0x3d5
	out	dx, al
	mov	al, 0x0e	;Index reg.: cursor pos (high byte)
	dec	edx		;0x3d4
	out	dx, al
	mov	al, ah		;Write high byte
	inc	edx		;0x3d5
	out	dx, al
	mpop	eax, edx
	ret


PAGE:		;; ( -- )  Clear the screen (or form-feed, if printing)
clear:
;
; Clear the screen
;
	mpush	eax, ecx, edx, edi
	mov	edi, [vidbuf]		;Destination
	mov	eax, [rows]
	mul	dword [cols]
	mov	ecx, eax		;Count = rows * cols
	mov	ah, [attr]		;attr/char to clear to
	xor	al, al
	rep	stosd

	mov	[cursor], word 0	;Position cursor at top left
	call	update_cursor
	mpop	eax, ecx, edx, edi
	ret
ePAGE


print_num:	;; ( num -- )
	DPUSH	0
	call	print_num_pad
	ret


print_num_pad:	;; ( num pad -- )
PRINT_NUM_PAD
;;
;; Paremeters:
;;	Number to print
;;	Padding width (0 = don't pad).  Set bit 7 to 0-pad.
;;
	pusha
	mov	eax, S(1)	; the number to print
	mov	ebx, [_BASE]
	xor	ecx, ecx

	or	eax, eax	; zero, positive, or negative?
	jz	.zero
	jg	.push
	DPUSH	'-'		; negative..
	call	putc
	neg	eax
.push				; push each digit, from right to left
	xor	edx, edx
	div	ebx
	or	eax, eax	; test to see if we're done
	jnz	.nz
	or	edx, edx
	jz	.print
.nz
	push	edx
	inc	ecx
	jmp short .push
.zero
	push	eax
	inc	ecx
.print
	;; Padding
	mov	edx, S(0)
	mov	al, 32		;space-pad by default
	test	dl, 0x80
	jz	.1
	mov	al, '0'		;zero-pad
	and	dl, ~0x80
.1
	cmp	ecx, edx	;Need to pad?
	jae	.loop
	sub	edx, ecx	;# of pad chars
	push	ecx
	mov	ecx, edx
.pad
	DPUSH	eax
	call	putc		;Print the pad characters
	loop	.pad
	pop	ecx
.loop
	;; Print the digits
	pop	eax
	add	al, '0'
	cmp	al, '9'
	jbe	.putc
	add	al, 7
.putc	DPUSH	eax
	call	putc
	loop	.loop
	popa
	add	ebp, 8		; 2DROP
	ret
ePRINT_NUM_PAD


md:
MD:	;; ( addr -- )
;
; Memory dump
;
	pusha
	mpush	[_BASE]
	mov dword [_BASE], 16

	DPOP	esi			;starting address
	mov	ecx, 16			;print 16 lines
	xor	eax, eax
.loop
	phex	esi			;; address
	space
	push	esi
	call	.8
	pchr	'-'
	space
	call	.8
	pop	esi
	call	.ascii
	cr
	loop	.loop

	mpop	[_BASE]
	popa
	add	ebp, 4		; DROP
	ret

.8	push	ecx
	mov	ecx, 8
.l8	lodsb
	DPUSH	eax
	DPUSH	0x82
	call	print_num_pad
	space
	loop	.l8
	pop	ecx
	ret

.ascii	push	ecx
	space
	mov	ecx, 16
.10	lodsb
	cmp	al, 26
	jbe	.11
	cmp	al, 127
	jb	.12
.11	mov	al, '.'
.12	pchr	eax
	loop	.10
	pop	ecx
	ret
eMD


AT_XY		;; ( x y -- )
	mpush	eax, edx
	DPOP	eax
	mul	dword [cols]
	DPOP	edx
	shl	dx, 1
	add	ax, dx
	mov	[cursor], ax
	call	update_cursor
	mpop	eax, edx
	ret
eAT_XY


	SECTION .data
vidbuf	dd	0xB8000
rows	dd	25
cols	dd	80*2
cursor	dw	0		;Linear cursor position
attr	db	0x07		;Color/attribute
dmode	db	0		;0=text, else graphics


	SECTION .text
FONT	;; ( -- addr )
	DPUSH	font
	ret
eFONT


GPUTC	;; ( char -- )
	mpush	eax, ecx, esi, edi
	mov	ecx, 8	; # of scanlines per character
	DPOP	esi
	shl	esi, 3	; *8  (# of scanlines per character)
	add	esi, font

	mov	ax, word [gcurs]
	inc	word [gcurs]
	xor	dx, dx
	div	word [gcols]
	movzx	edi, dx
	shl	edi, 3	; *8  (character width)
	mul	word [glsize]
	movzx	eax, ax
	add	edi, eax
	add	edi, [fbuf]
.v	mov	ah, [esi]
	inc	esi
	push	ecx
	mov	ecx, 8	; char. width
.h	xor	al, al	; 0 = black
	test	ah, 0x80
	jz	.0
	mov	al, 7	; white
.0	stosb
	shl	ah, 1
	loop	.h
	pop	ecx
	add	edi, 320-8	; screen width
	loop	.v
	mpop	eax, ecx, esi, edi
	ret
eGPUTC


DELAY	;; ( count -- )
;;	count = Number of 15.085-microsecond cycles to delay
;;
	mpush	eax, ecx
	DPOP	ecx
	in	al, 0x61
	and	al, 0x10
	mov	ah, al
.l1	in	al, 0x61
	and	al, 0x10
	cmp	al, ah
	jz	.l1
	mov	ah, al
	loop	.l1
	mpop	eax, ecx
	ret
eDELAY


	SECTION .data
font:	incbin "../fonts/gr8x8.psf"
gcurs	dw	0
fbuf	dd	0xA0000		; framebuffer
gcols	dw	40
glsize	dw	320*8		; bytes per line
