; gdt.inc - Macros for building the GDT and IDT
;
; Copyright (C) 1999 Thomas Novelli
;	Retro is distributed under the terms of the GNU LGPL, version 2 or
;	greater. (For the text of the license, see the file COPYING.LIB)
;

;Each descriptor should have exactly one of next 8 codes to define the type
;of descriptor
D_LDT           EQU       02h   ;LDT segment
D_TASK          EQU       05h   ;Task gate
D_TSS           EQU       09h   ;TSS
D_CALL          EQU       0Ch   ;386 call gate
D_INT           EQU       0Eh   ;386 interrupt gate
D_TRAP          EQU       0Fh   ;386 trap gate
D_DATA          EQU       10h   ;Data segment
D_CODE          EQU       18h   ;Code segment

;Descriptors may include the following as appropriate:
D_DPL3          EQU       60h   ;DPL3 or mask for DPL
D_DPL2          EQU       40h
D_DPL1          EQU       20h
D_PRESENT       EQU       80h   ;Present
D_NOT_PRESENT   EQU       80h   ;Not Present
                                ;Note, the PRESENT bit is set by default
                                ;Include NOT_PRESENT to turn it off
                                ;Do not specify D_PRESENT

;Segment descriptors (not gates) may include:
D_ACC           EQU       01h   ;Accessed (Data or Code)

D_WRITE         EQU       02h   ;Writable (Data segments only)
D_READ          EQU       02h   ;Readable (Code segments only)
D_BUSY          EQU       02h   ;Busy (TSS only)

D_EXDOWN        EQU       04h   ;Expand down (Data segments only)
D_CONFORM       EQU       04h   ;Conforming (Code segments only)

D_BIG           EQU     4000h   ;Default to 32 bit mode (USE32)
D_BIG_LIM       EQU     8000h   ;Limit is in 4K units (page granular)



%macro start_gdt 0
%push table
%$startoftable
	dw	%$limitoftable
	dd	%$startoftable
	dw	0
%endmacro

%macro end_gdt 0
%$limitoftable	equ	$-%$startoftable
%pop
%endmacro

%macro desc 3	;Usage: selector-label   desc   base, limit, flags
%00	equ $-%$startoftable
	dd	(%2 & 0xFFFF) + ((%1 & 0xFFFF) << 16)
	dd	(%1 & 0xFF000000) + (%2 & 0xF0000) + (((%3) ^ D_PRESENT) << 8) + ((%1 >> 16) & 0xFF)
%endmacro


;Interrupt Descriptor Table stuff
;
;Allowed types: D_TASK, D_INT, and D_TRAP
;Other flags allowed: D_DPL*, D_PRESENT, D_NOT_PRESENT
;
%macro idesc 3	;Usage: idesc  selector, offset, flags
	dw	((%2-$$) & 0xFFFF)
	dw	%1
	db	0
	db	((%3) ^ D_PRESENT)
	dw	((%2-$$) >> 16)
%endmacro
