; mm.inc - memory management
;
; Typical MM interface:
;	ALLOC ( size -- addr)
;	FREE ( addr -- )
;	REALLOC ( size addr1 -- addr1 )
;
; Here's the first method I thought up: Put the list of free physical pages
; on the pages themselves.  The first word of each free page will be a
; pointer to the next free page (or 0, for the last page).  To allocate
; pages, you just go to a free page, map it, move the "first page" pointer
; to the next free page, and repeat for as many pages as you need.  To free
; pages, you link each page into the freelist and unmap it.
;
; Time considerations:  The time to free pages is directly proportional to
; the number of pages (linear time).  Allocation is also linear, but
; slightly slower, and with an initial delay needed to allocate logical
; address space.
;
; * Logical address allocation:  We need a freelist, Lfree (unordered) and
; Lused (ordered by logical address).  For allocation, just do a first-fit,
; split the block, remove it from Lfree and insert it in Lused.  For
; deallocation, search Lused for the address.  A tree would probably become
; unbalanced quickly, so we should use a hash table (any other ideas?), or a
; linked list to start with.
; 
	struc Llist
.addr	resd	1
.size	resd	1
.next	resd	1
	endstruc
;
; When a block is deallocated, link its Llist entry into a list of unused
; Llist structs.


	SECTION .text
pagefault:
	iret


init_phys_pages:
; We'll need to change this later to grab all the free areas of memory,
; skipping around the GDT/IDT, page tables, kernel, video framebuffers,
; ROMs, etc.
	mov	edi, [Pfree.first]
	mov	ecx, 0x100
.loop	


alloc:	;; ( size -- addr )
	mpush	eax, ecx, edx, edi
	; allocate a logical address
	; edi = PTE of starting logical address
	; ecx = # of pages to allocate
	; dx = PTE flags (bits 0-11)
	; Then, map the pages:
	DPUSH	edi
	mov	dx, 0x02	;PTE flags: writable
	mov	eax, Pfree.first
.loop	mov	eax, [eax]	;get next address
	or	eax, eax	;no free memory?
	jz	.SOL
	or	ax, dx		;mask in PTE flags
	stosd			;store PTE
	loop	.loop
.end	mpop	eax, ecx, edx, edi
	ret
.SOL	mov	S(0), 0		;Return 0 (error)
	jmp short .end


	SECTION .data
Pfree				;free Physical pages
.first	dd	0x300000	;First unused page (borrow from here)
.last	dd	0x30F000	;Last unused page (return here)
