; page.asm - Memory paging
;
; Copyright (C) 1999 Thomas Novelli
;	Retro is distributed under the terms of the GNU LGPL, version 2 or
;	greater. (For the text of the license, see the file COPYING.LIB)
;
; Progress:
;	3/22/99 Well, it works!
;
;	4/9/99	Added a simple page fault handler that allocates pages for
;		kernel modules & stack as needed.  Moved page table init
;		code to pmode.inc (loader.inc needs it there).
;
; To do:
;
;
; Notes:
;	Map the first 4M directly to physical memory; that way it won't
;	upset anything I did before paging.
;
;	TLB must be flushed any time the page tables are changed, by
;	reloading CR3 (MOV EAX, CR3; MOV CR3, EAX). This happens
;	automatically during a task switch, however, so no need to worry
;	about that.
;

%include "memory.inc"
%include "boot.mac"

extern	print_hex, putc, print, patch_int


	SEGMENT .text

init:	callproc print, msg_init
	callproc patch_int,	14, page_fault
	ret



clear_page:
;
; Usage:
;	mov	edi, <page address>
;	call	clear_page
;
	xor	eax, eax
	mov	ecx, 0x400
	cld
	rep	stosd
	ret



_malloc:
;
; Parameters: length
;
	ret


_malloca:
;
; Allocate aligned memory
;
; Parameters:
;	length
;	alignment
;
	ret



	global mcopy
mcopy:
;
; Copy a region of memory to another location
;
; Parameters: source, dest, length
;
	mpush	eax, ecx, esi, edi
	mov	esi, [esp+7*4]
	mov	edi, [esp+6*4]
	mov	ecx, [esp+5*4]
	cld
	rep	movsb
	mpop	eax, ecx, esi, edi
	ret



page_fault:
;;
;; Stack frame:
;;	Error code (bit 0=protection fault, bit 1=write, bit 2=user mode)
;;	EIP
;;	CS
;;	EFLAGS
;;
	mpush	eax, ebx
	mov	eax, CR2	; address that caused fault
	mov	ebx, [esp+2*4]	; error code
;; If in supervisor mode, page not present fault, and CR2 > FFC00000 (in
;; last 4 megs), then just allocate the page.
	test	ebx, 5		; usermode or protection fault?
	jnz	.crap
	cmp	eax, 0xFFC00000	; outside kernel area?
	jb	.crap

	mov	ebx, [freeptr]		;Allocate
	shr	eax, 10
	and	eax, 0xFFC
	mov	[PAGE_TBL+0x2000+eax], ebx
	add dword [freeptr], 0x1000
	mov	eax, CR3	; Flush TLB
	mov	CR3, eax
	mpop	eax, ebx
	add	esp, 4		; drop error code
	sti
	iretd

.crap	;; We won't deal with this just yet :)
	callproc print,	pf_msg
	callproc print_hex, eax
	callproc print,	eip_msg
	callproc print_hex, [esp+3*4]	;EIP
	callproc print,	ec_msg
	callproc print_hex, ebx		;error code
	cr
	sti
	jmp short $



	SECTION .data
msg_init	db	"++ Paging",0
pf_msg	db	10," * EXC14: Page Fault",10,"CR2 = ",0
eip_msg	db	10,"EIP = ",0
ec_msg	db	10," ec = ",0

freeptr		dd	0x103007	;First page of free physical RAM
